#define G_LOG_DOMAIN "boilerplate-preferences-view"

#include "config.h"

#include "boilerplate-preferences-view.h"
#include "boilerplate-preferences-page.h"
#include "boilerplate-preferences-group.h"

#include <gtk/gtk.h>

/**
 * SECTION:boilerplate-preferences-view2
 * @title: BoilerplatePreferencesView
 * @short_description: A #GtkBin instance created for #BoilerplatePreferences
 * @include: boilerplate-preferences-view.h
 * @see_also: #GtkBin, #BoilerplatePreferencesView
 *
 * The #BoilerplatePreferencesView is an instance #GtkBin, it's the base object
 * for the preferences options. It contains a #GtkStack and manages the adding,
 * removing, and update of options in the preferences view. This allows for the
 * back-end code for managing the preferences ui to be added to either a
 * #GtkWindow, #GtkDialog or other custom #GtkWidget without re-writing an
 * entire GObject.
 **/

typedef struct _BoilerplatePreferencesViewPrivate
{
    GHashTable *pages;
    GHashTable *groups;
    GHashTable *options;
} BoilerplatePreferencesViewPrivate;

struct _BoilerplatePreferencesView
{
    GtkBin parent_instance;

    GtkWidget *box;
    GtkWidget *sidebar;
    GtkWidget *pages;
};

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplatePreferencesView, boilerplate_preferences_view, GTK_TYPE_BIN)

/**
 * boilerplate_preferences_view_add_page:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Id of the page for internal usage
 * @page_title: The title that will apear in the #GtkStackSidebar
 *
 * Adds a new page to the #GtkStack of pages.
 */
void
boilerplate_preferences_view_add_page (BoilerplatePreferencesView *self,
                                       const gchar                *page_name,
                                       const gchar                *page_title)
{
    BoilerplatePreferencesPage *page;
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);

    g_assert (BOILERPLATE_IS_PREFERENCES_VIEW (self));
    g_assert (page_name != NULL);
    g_assert (page_title != NULL);

    page = boilerplate_preferences_page_new ();

    if (page == NULL)
    {
        g_warning ("New page could not be created.");
        return;
    }

    if (page_name == NULL)
    {
        g_warning ("Page name can not be NULL.");
        return;
    }

    if (page_title == NULL)
    {
        g_warning ("Page title can not be NULL.");
        return;
    }

    gtk_container_add_with_properties (GTK_CONTAINER (self->pages), GTK_WIDGET (page),
                                       "name", page_name,
                                       "title", page_title,
                                       NULL);

    g_hash_table_insert (priv->pages, g_strdup (page_name), page);
}

/**
 * boilerplate_preferences_view_add_page_with_child:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Id of the page for internalg_strdup usage
 * @page_title: The title that will apear in the #GtkStackSidebar
 * @child: A #GtkWidget to add to the new page
 *
 * Adds a child #GtkWidget to a new page, then adds it to #GtkStack of pages.
 */
void
boilerplate_preferences_view_add_page_with_child (BoilerplatePreferencesView *self,
                                                  const gchar                *page_name,
                                                  const gchar                *page_title,
                                                  GtkWidget                  *child)
{
    GtkWidget *page;

    g_assert (BOILERPLATE_IS_PREFERENCES_VIEW (self));
    g_assert (page_name != NULL);
    g_assert (page_title != NULL);
    g_assert (GTK_IS_WIDGET (child));

    boilerplate_preferences_view_add_page (self, page_name, page_title);

    page = boilerplate_preferences_view_get_page (self, page_name);

    if (page == NULL)
    {
        g_warning ("No page named \"%s\" could be found.", page_name);
        return;
    }

    gtk_container_add (GTK_CONTAINER (page), GTK_WIDGET (child));
}

/**
 * boilerplate_preferences_view_add_child_to_page:
 * @object: An instance of #BoilerplatePreferencesView
 * @page_name: Id of the page for internal usage
 * @child: A #GtkWidget to add to the new page
 *
 * Sets a @child into a page with @name.
 */
void
boilerplate_preferences_view_add_child_to_page (BoilerplatePreferencesView *self,
                                                const gchar                *page_name,
                                                GtkWidget                  *child)
{
    GtkWidget *page;

    g_assert (BOILERPLATE_IS_PREFERENCES_VIEW (self));
    g_assert (page_name != NULL);
    g_assert (GTK_IS_WIDGET (child));

    page = boilerplate_preferences_view_get_page (self, page_name);

    if (page == NULL)
    {
        g_warning ("No page named \"%s\" could be found.", page_name);
        return;
    }

    gtk_container_add (GTK_CONTAINER (page), GTK_WIDGET (child));
}

/**
 * boilerplate_preferences_view_add_child_to_page:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Id of the page for internal usage
 * @group_name: The id of the page for internal usage
 * @group_title: The name to be shown above the group
 *
 * Creates a group with a title.
 */
void
boilerplate_preferences_view_add_group (BoilerplatePreferencesView *self,
                                        const gchar                *page_name,
                                        const gchar                *group_name,
                                        const gchar                *group_title)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);
    GtkWidget *page;
    BoilerplatePreferencesGroup *group;

    g_assert (BOILERPLATE_IS_PREFERENCES_VIEW (self));
    g_assert (page_name != NULL);
    g_assert (group_name != NULL);
    g_assert (group_title != NULL);

    page = boilerplate_preferences_view_get_page (self, page_name);

    if (page == NULL)
    {
        g_warning ("No page named \"%s\" could be found.", page_name);
        return;
    }

    if (group_name == NULL)
    {
        g_warning ("Group name can not be NULL.");
        return;
    }

    if (group_title == NULL)
    {
        g_warning ("Group title can not be NULL.");
        return;
    }

    group = g_object_new (BOILERPLATE_TYPE_PREFERENCES_GROUP,
                          "name", group_name,
                          "title", group_title,
                          NULL);

    gtk_container_add (GTK_CONTAINER (page), GTK_WIDGET (group));

    g_hash_table_insert (priv->groups, g_strdup_printf ("%s.%s", page_name, group_name), group);
}

/**
 * boilerplate_preferences_view_add_child_to_page:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Id of the page for internal usage
 * @group_name: The id of the page for internal usage
 * @widget: #GtkWidget to add to @group_name in @page_name
 *
 * Adds an option to
 */
void
boilerplate_preferences_view_add_option_row (BoilerplatePreferencesView *self,
                                             const gchar                *page_name,
                                             const gchar                *group_name,
                                             GtkWidget                  *widget)
{
    GtkWidget *page;
    GtkWidget *group;
    GtkWidget *separator;

    g_assert (BOILERPLATE_IS_PREFERENCES_VIEW (self));
    g_assert (page_name != NULL);
    g_assert (group_name != NULL);
    g_assert (GTK_IS_WIDGET (widget) && widget != NULL);

    page = boilerplate_preferences_view_get_page (self, page_name);

    if (page == NULL)
    {
        g_warning ("No page named \"%s\" could be found.", page_name);
        return;
    }

    if (group_name == NULL)
    {
        g_warning ("Group name can not be NULL.");
        return;
    }

    group = boilerplate_preferences_view_get_group (self, page_name, group_name);

    if (widget == NULL)
    {
        g_warning ("No widget to add to option.");
        return;
    }

    boilerplate_preferences_group_add (BOILERPLATE_PREFERENCES_GROUP (group), GTK_WIDGET (widget));

    separator = g_object_new (GTK_TYPE_SEPARATOR,
                              "visible", TRUE,
                              "can_focus", FALSE,
                              "orientation", GTK_ORIENTATION_HORIZONTAL,
                              NULL);

    boilerplate_preferences_group_add (BOILERPLATE_PREFERENCES_GROUP (group), GTK_WIDGET (separator));
}

/**
 * boilerplate_preferences_view_get_page:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Name of the page to search for
 *
 * Gets the a #BoilerplatePreferencesPage with @page_name.
 *
 * Returns: (nullable): Returns a #GtkWidget of the page with @page_name.
 */
GtkWidget *
boilerplate_preferences_view_get_page (BoilerplatePreferencesView *self,
                                       const gchar                *page_name)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);
    GtkWidget *page;

    g_return_val_if_fail (BOILERPLATE_IS_PREFERENCES_VIEW (self), NULL);
    g_return_val_if_fail (page_name != NULL, NULL);

    page = g_hash_table_lookup (priv->pages, g_strdup (page_name));

    if (page == NULL)
    {
        g_warning ("No page named \"%s\" in pages hash table.", page_name);
        return NULL;
    }

    return boilerplate_preferences_page_get_box (BOILERPLATE_PREFERENCES_PAGE (page));

/*
    page = gtk_stack_get_child_by_name (GTK_STACK (self->pages), page_name);

    if (page == NULL)
    {
        return NULL;
    }

    return boilerplate_preferences_page_get_box (BOILERPLATE_PREFERENCES_PAGE (page));
*/
}

/**
 * boilerplate_preferences_view_get_group:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Name of the page to search for
 * @group_name: Name of the group to search for
 *
 * Gets the a #BoilerplatePreferencesGroup with @group_name in @group_page.
 *
 * Returns: (nullable): Returns a #GtkWidget of the page with @group_name.
 */
GtkWidget *
boilerplate_preferences_view_get_group (BoilerplatePreferencesView *self,
                                        const gchar                *page_name,
                                        const gchar                *group_name)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);
    GtkWidget *group;

    g_return_val_if_fail (BOILERPLATE_IS_PREFERENCES_VIEW (self), NULL);
    g_return_val_if_fail (group_name != NULL, NULL);

    g_message ("%s", g_strdup_printf ("%s.%s", page_name, group_name));

    group = g_hash_table_lookup (priv->groups, g_strdup_printf ("%s.%s", page_name, group_name));

    if (group == NULL)
    {
        g_warning ("No group named \"%s\" in pages hash table.", group_name);
        return NULL;
    }

    return group;
}

/**
 * boilerplate_preferences_view_add_switch:
 * @self: An instance of #BoilerplatePreferencesView
 * @page_name: Name of the page to search for
 * @group_name: Name of the group to search for
 * @schema: #GSetting schema path
 * @key: #GSetting key name to search for
 * @name: Name of the option
 * @description: Short description of the option
 *
 * Adds a switch option
 */
void
boilerplate_preferences_view_add_switch (BoilerplatePreferencesView *self,
                                         const gchar                *page_name,
                                         const gchar                *group_name,
                                         const gchar                *schema,
                                         const gchar                *key,
                                         const gchar                *name,
                                         const gchar                *description)
{

}

/*
 * boilerplate_preferences_view_finalize:
 * @object: An instance of #BoilerplatePreferencesView
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
boilerplate_preferences_view_finalize (GObject *object)
{
    BoilerplatePreferencesView *self = BOILERPLATE_PREFERENCES_VIEW (object);
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);

    g_clear_pointer (&priv->pages, g_hash_table_unref);
    g_clear_pointer (&priv->groups, g_hash_table_unref);
    g_clear_pointer (&priv->options, g_hash_table_unref);

    G_OBJECT_CLASS (boilerplate_preferences_view_parent_class)->finalize (object);
}

/*
 * boilerplate_preferences_view_init:
 * @self: An instance of #BoilerplatePreferencesView
 *
 * Initializes various aspects of the window. Anything the window would need
 * before becoming presented should be done here.
 */
static void
boilerplate_preferences_view_init (BoilerplatePreferencesView *self)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);

    gtk_widget_init_template (GTK_WIDGET (self));

    priv->pages = g_hash_table_new (g_str_hash, g_str_equal);
    priv->groups = g_hash_table_new (g_str_hash, g_str_equal);
    priv->options = g_hash_table_new (g_str_hash, g_str_equal);
}

/*
 * boilerplate_preferences_view_class_init:
 * @klass: An instance of #BoilerplatePreferencesViewClass
 *
 * Initializes the various virtural functions in #BoilerplatePreferencesViewClass.
 * In addtition it connects the various #GtkWidgets from the gresource file
 * using gtk_widget_class_set_template_from_resource().
 */
static void
boilerplate_preferences_view_class_init (BoilerplatePreferencesViewClass *klass)
{
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = boilerplate_preferences_view_finalize;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/preferences/view.ui");

    gtk_widget_class_bind_template_child (widget_class, BoilerplatePreferencesView, box);
    gtk_widget_class_bind_template_child (widget_class, BoilerplatePreferencesView, sidebar);
    gtk_widget_class_bind_template_child (widget_class, BoilerplatePreferencesView, pages);
}

/**
 * boilerplate_preferences_view_new:
 *
 * Creates a new #BoilerplatePreferencesView, a sub-object of #GtkBin.
 *
 * ## Usage
 * |[
 * BoilerplatePreferencesView *preferences_view;
 *
 * preferences_view = boilerplate_preferences_view_new ();
 * ]|
 *
 * Returns: #BoilerplatePreferencesView; a sub-object of #GtkBin
 */
BoilerplatePreferencesView *
boilerplate_preferences_view_new (void)
{
    return g_object_new (BOILERPLATE_TYPE_PREFERENCES_VIEW,
                         NULL);
}

